# -*- coding: utf-8 -*-

import video
import subprocess
import os
import os.path
import logging

class VideoCutter(object):
  import multiprocessing as mp
  import video._batch_cut
  
  """Implements functionality to cut a video into segments"""
  
  def __init__(self, videolist, encoder, ffprobe='/usr/bin/ffprobe', tmpdir='.'):
    """Constructs a new cutter on one or more underlying files.
       video is either a single filename, or a list of filenames
       that specify consecutive segments of the video.
    """
    super(VideoCutter, self).__init__()
    self.video = video.Video(videolist, ffprobe=ffprobe)
    self.encoder = encoder
    self.tmpdir = tmpdir
    self.frame_length = int(1000.0 / self.video.get_fps()) + 1
  
  def get_video(self):
    return self.video
  
  def cut(self, start, end, dest_filename, overwrite=False, clip=True, min_length=None):
    """Cuts a video segment from start to end, saving the
       result in dest_filename.
       If overwrite is true, deletes the destination filename
       before cutting.
    """
    if overwrite:
      try:
        os.remove(dest_filename)
      except:
        pass
    try:
      (video_file, vid_start, vid_end) = self.video.get_region(start, end, clip)
      if min_length is None or end - start >= min_length:
        self.encoder.transcode(video_file, dest_filename,
                               start=vid_start, end=vid_end)
      else:
        # Target video is too short, need to stretch
        # make sure we have at least one frame
        vid_end = max(vid_end, vid_start + self.frame_length)
        videodir = self._extract_frames(video_file, dest_filename, vid_start, vid_end)
        stretch_frames(videodir, self.video.get_fps(), min_length)
        self.encoder.transcode(os.path.join(videodir, "%05d.ppm"), dest_filename,
                               fps=self.video.get_fps())
        self._clear_dir(videodir)
    except video.RegionStraddlesSegmentsException:
      # if a region straddles two videos, we have no choice but to save all the
      # frames and transcode them from scratch. Sucks, but what else is there to do?
      log = logging.getLogger("video")
      log.warn("(%d, %d) straddles multiple files, falling back to extracting frames as images" % (start, end))
      (video1, vid_start) = self.video.get_segment_and_offset(start)
      (video2, vid_end) = self.video.get_segment_and_offset(end)
      tmpdir1 = self._extract_frames(video1, dest_filename, vid_start, None)
      tmpdir2 = self._extract_frames(video2, dest_filename, 0, vid_end)
      # Both sequences start with counts from 00001.ppm. Find highest frame # of first sequence
      # and rename + move second sequence accordingly
      first_files = sorted(os.listdir(tmpdir1))
      if len(first_files) > 0:
        last_file = os.path.basename(first_files[-1])
        (next_in_seq, ext) = os.path.splitext(last_file)
      else:
        next_in_seq = 0 # first directory has zero files
      next_in_seq = int(next_in_seq) + 1
      for file in sorted(os.listdir(tmpdir2)):
        os.rename(os.path.join(tmpdir2, file),
                  os.path.join(tmpdir1, "%05d.ppm" % next_in_seq))
        next_in_seq += 1
      os.rmdir(tmpdir2)
      # Do we need to stretch?
      if min_length is not None and end - start < min_length:
        stretch_frames(tmpdir1, self.video.get_fps(), min_length)
      # make target video. need to specify framerate manually
      self.encoder.transcode(os.path.join(tmpdir1, "%05d.ppm"), dest_filename,
                             fps=self.video.get_fps())
      # clean up temp files
      self._clear_dir(tmpdir1)
        
  
  def batch_cut(self, videolist, num_processes=mp.cpu_count(), overwrite=False):
    """Batch-processes a number of videos.
       video_cutter is an instance of VideoCutter.
       videolist contains the videos to process, in the tuple format
         (start in ms, end in ms, destination filename, min_length)
       num_processes is the number of processes to use (default: cpu count on system)
       overwrite, if true, causes the destination to be overwritten if it already exists
       Returns true if all cuts succeeded
    """
    video._batch_cut.batch_process_video(self, videolist, num_processes, overwrite)

  def _extract_frames(self, video, dest, start, end):
    frame_grabber = FrameGrabber('all', 'ppm')
    tmpdir = os.path.join(self.tmpdir, os.path.basename(video) + os.path.basename(dest))
    os.mkdir(tmpdir)
    frame_grabber.transcode(video, os.path.join(tmpdir, "%05d.ppm"), start, end)
    return tmpdir
  
  def _clear_dir(self, tmpdir):
    for file in os.listdir(tmpdir):
      os.remove(os.path.join(tmpdir, file))
    os.rmdir(tmpdir)


class Encoder(object):
  """Represents a video encoder"""
  # we may want to add some common capabilities here, but not now
  pass


class FFMPEGEncoder(Encoder):
  """represents an ffmpeg-based encoder"""
  def __init__(self, ffmpeg='/usr/bin/ffmpeg', args=[], **kw):
    super(FFMPEGEncoder, self).__init__()
    self.ffmpeg = ffmpeg
    self.args = args
    
  def transcode(self, source, destination, start=None, end=None, fps=None):
    """Transcodes a video.
       if start and end are specified, denote the start and end time in milliseconds
    """
    command = self._make_commandline(source, destination, start, end, fps)
    print command
    log = logging.getLogger("video")
    log.debug(" ".join(command))
    ffmpeg_process = subprocess.Popen(command,
                                      stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE, close_fds=True)
    (stdout, stderr) = ffmpeg_process.communicate()
    if ffmpeg_process.returncode != 0:
      raise FFMPEGError(ffmpeg_process.returncode, stderr)
  
  def arguments(self):
    """Returns additional commandline arguments for ffmpeg"""
    return [a for a in self.args] # need to copy to prevent in-place overwriting by subclasses
  
  def _make_timecode_hms(self, timecode):
    # Converts milliseconds to hours, minutes, seconds, milliseconds
    (seconds, milliseconds) = divmod(timecode, 1000)
    (minutes, seconds) = divmod(seconds, 60)
    (hours, minutes) = divmod(minutes, 60)
    return "%02d:%02d:%02d.%03d" % (hours, minutes, seconds, milliseconds)
  
  def _make_commandline(self, source, destination, start, end, fps=None):
    cmd = [self.ffmpeg]
    if fps is not None:
      cmd.extend(['-r', str(fps)])
    cmd.extend(['-i', source])
    if start is not None:
      cmd.extend(['-ss', self._make_timecode_hms(start)])
    if start is not None and end is not None:
      cmd.extend(['-t',  self._make_timecode_hms(end - start)])
    cmd.extend(self.arguments())
    cmd.append(destination)
    return cmd


class X264Encoder(FFMPEGEncoder):
  """Encodes H.264 files via libx264"""
  def __init__(self, quality=23, bitrate=None, preset='slow', **kw):
    super(X264Encoder, self).__init__(**kw)
    self.quality = quality
    self.bitrate = bitrate
    self.preset = preset
  
  def arguments(self):
    args = super(X264Encoder, self).arguments()
    args.extend(["-vcodec", "libx264"])
    if self.quality is not None:
      args.extend(["-crf", str(self.quality)])
    if self.bitrate is not None:
      args.extend(["-b", self.bitrate])
    if self.preset is not None:
      args.extend(["-vpre", self.preset])
    return args



class PCMEncoder(FFMPEGEncoder):
  """Encodes PCM files"""
  def __init__(self, codec="pcm_s16le", **kw):
    super(PCMEncoder, self).__init__(**kw)
    self.codec = codec
  
  def arguments(self):
    args = super(PCMEncoder, self).arguments();
    args.extend(["-vn", "-acodec", self.codec])
    return args


class Mpeg4Encoder(FFMPEGEncoder):
  """Encodes Mpeg4 files"""
  def __init__(self, bitrate=None, **kw):
    super(Mpeg4Encoder, self).__init__(**kw)
    self.bitrate = bitrate
  
  def arguments(self):
    args = super(Mpeg4Encoder, self).arguments()
    args.extend(["-vcodec", "mpeg4"])
    if self.bitrate is not None:
      args.extend(["-b", self.bitrate])
    return args
  

class FrameGrabber(FFMPEGEncoder):
  """Grabs a one or more frames from a video"""
  valid_types = set(["start", "middle", "all"])
  
  def __init__(self, type='start', format='jpg', **kw):
    """Initializes the frame grabber.
        type is one of 'start', 'middle'
        format is an image format that ffmpeg understands.
    """
    super(FrameGrabber, self).__init__(**kw)
    if type in self.valid_types:
      self.type = type
    else:
      raise InvalidGrabTypeException(type)
    self.format = format
    
  def transcode(self, source, destination, start=None, end=None, **kw):
    if self.type == 'start':
      return super(FrameGrabber, self).transcode(source, destination, start, end=None, **kw)
    elif self.type == 'middle':
      if start is None or end is None:
        raise MissingTimecodeException()
      center = (start + end) / 2
      return super(FrameGrabber, self).transcode(source, destination, center, end=None, **kw)
    elif self.type == 'all':
      return super(FrameGrabber, self).transcode(source, destination, start, end, **kw)
    
  def arguments(self):
    args = super(FrameGrabber, self).arguments()
    if self.type != 'all':
      args.extend(['-vframes', '1'])
    args.extend(['-f', 'image2'])
    return args


class FFMPEGError(Exception):
  """Exception that is raised if ffmpeg encounters an error"""
  def __init__(self, retcode, text):
    super(FFMPEGError, self).__init__("ffmpeg returned error %d; %s" % (retcode, text))


class InvalidGrabTypeException(Exception):
  """Exception that is raised if an invalid type for frame grabbing
     is specified.
  """
  def __init__(self, type):
    super(InvalidGrabTypeException, self).__init__("Invalid grab type: %s" % type)


class MissingTimecodeException(Exception):
  """Exception that is raised if frame grabbing requires timecodes
  """
  def __init__(self, type):
    super(MissingTimecodeException, self).__init__("Grabbing middle frame needs timecodes")


def stretch_frames(videodir, fps, duration, append_only=False):
  """Pads a set of image frames, such that the total video
     is as long as the specified duration.
     
     videodir is the directory containing the frames of the form nnnnn.ppm
     fps is the target frames per second
     duration is the target duration
     append_only if true, append frames only to end, otherwise pad equally at
                   start and end
  """
  num_frames = len(os.listdir(videodir))
  min_num_frames = int(round(duration * fps / 1000.0))
  if min_num_frames > num_frames:
    missing_frames = min_num_frames - num_frames
    if append_only:
      start_frames = 0
      end_frames = missing_frames
    else:
      start_frames = (missing_frames + 1) / 2
      end_frames = missing_frames / 2
    # shift frames to make space for start frames
    if  start_frames > 0:
      for i in xrange(num_frames, 0, -1):
        old_name = "%05d.ppm" % i
        new_name = "%05d.ppm" % (i + start_frames)
        os.rename(os.path.join(videodir, old_name), os.path.join(videodir, new_name))
      # copy start frames via symlink
      first_frame = "%05d.ppm" % (start_frames + 1)
      for i in xrange(1, start_frames + 1):
        target_frame = "%05d.ppm" % i
        os.symlink(os.path.abspath(os.path.join(videodir, first_frame)), os.path.join(videodir, target_frame))
    # copy end frames via symlink
    last_frame = "%05d.ppm" % (num_frames + start_frames) 
    for i in xrange(1, end_frames + 1):
      target_frame = "%05d.ppm" % (i + num_frames + start_frames)
      os.symlink(os.path.abspath(os.path.join(videodir, last_frame)), os.path.join(videodir, target_frame))
