package edu.bu.signstream.common.util;

public class CoordinateTimeConverter {
    // How many frames should the max zoom level show per width
    private static final long MAX_ZOOM_FRAMES = 30;

    // When fitting the zoom, pad the width by this amount.
    // The drawing code has a small margin, so fitting the exact width would cause
    // utterances to be slightly larger than the available drawing space.
    // This should really be accounted for in the drawing logic.
    private static final int WIDTH_PAD = -100;

    // Functions for remapping doubles from [0, 1] into a log scale from [0, 1].
    // LOG_A = 0.99 makes the value increase slowly near 0, and quickly near 1.
    // See https://www.desmos.com/calculator/ixdtn1t1ap
    private static final double LOG_A = 0.99;
    private static double logScale(double fr) {
        return Math.log(1 - LOG_A * fr) / Math.log(1 - LOG_A);
    }
    private static double invLogScale(double fr) {
        return (1 - Math.pow(1 - LOG_A, fr)) / LOG_A;
    }

    // Functions for remapping doubles from [0, 1] into [a, b]
    private static double lerp(double a, double b, double fr) {
        return (a * (1.0 - fr)) + (b * fr);
    }
    private static double invLerp(double a, double b, double v) {
        return (v - a) / (b - a);
    }

    // Dumb random defaults since the code calls this a million times before it should.
    // TODO: Remove these.
    private long timestep = 1001;
    private long duration = timestep * 30 * 60;
    private long width = 1000;

    private double zoom = 0.0;

    public void setTimeScale(long timestep, long duration) {
        this.timestep = timestep;
        this.duration = duration;
    }

    public void resize(long width) {
        this.width = width + WIDTH_PAD;
    }

    private double minFrameWidth() {
        long frames = duration / timestep;
        return (double)width / frames;
    }

    private double maxFrameWidth() {
        return (double)width / MAX_ZOOM_FRAMES;
    }

    private double frameWidth() {
        return lerp(minFrameWidth(), maxFrameWidth(), zoom);
    }

    public void fit(long duration, double width) {
        long frames = duration / timestep;
        this.zoom = invLerp(minFrameWidth(), maxFrameWidth(), (width + WIDTH_PAD) / frames);
    }

    public void fit(double width) {
        fit(duration, width);
    }

    public void zoom(double zoom) {
        this.zoom = logScale(zoom);
    }

    public double zoom() {
        return invLogScale(zoom);
    }

    public int convertTimeToCoordinate(int time) {
        long frame = time / timestep;
        return (int)Math.round(frame * frameWidth());
    }

	public int convertCoordinateToTime(int coordinate) {
        long frame = Math.round(coordinate / frameWidth());
        return (int)(frame * timestep);
	}

	/*
	    TODO: REMOVE EVERYTHING BELOW EVENTUALLY
	 */

    @Deprecated
    public int getMovieFrameDuration() {
        return (int)timestep;
    }

    @Deprecated
    public double getPixelsInTimeFrame() {
        return frameWidth();
    }

    @Deprecated
    public int getScreenWidth(int _duration) {
        return (int)width;
    }

    @Deprecated
    public int getAdjustedMovieTime(int coordinate) {
        return convertCoordinateToTime(coordinate);
    }

    @Deprecated
    public int getFrameNumber(int time) {
        return (int)(time / timestep);
    }

    @Deprecated
    public int getCorrespondingMovieTime(int frame) {
        return (int)(frame * timestep);
    }
}
