package edu.bu.signstream.media.gstreamer;

import org.freedesktop.gstreamer.*;
import org.freedesktop.gstreamer.elements.AppSink;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URI;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

public class GstAudioReader {
    public static final int RATE = 44100;

    private final CountDownLatch latch = new CountDownLatch(1);
    private final Pipeline pipeline = new Pipeline();

    protected ByteArrayOutputStream data = new ByteArrayOutputStream();

    public static FloatBuffer read(URI uri) {
        return new GstAudioReader(uri).await();
    }

    private GstAudioReader(URI uri) {
        var src = ElementFactory.make("uridecodebin", "decode");
        src.set("uri", uri.toString());

        var resample = ElementFactory.make("audioresample", "resample");
        var convert = ElementFactory.make("audioconvert", "convert");

        var sink = new AppSink("sink");
        sink.setCaps(new Caps("audio/x-raw,format=F32LE,layout=non-interleaved,channels=1,rate=" + RATE));

        sink.set("emit-signals", true);
        sink.connect((AppSink.NEW_SAMPLE) s -> {
            var sample = sink.pullSample();

            // Map the sample and copy it to the output stream
            var buf = sample.getBuffer();
            var view = buf.map(false);
            try {
                var b = new byte[view.remaining()];
                view.get(b);
                data.write(b);
            } catch (IOException e) {
                e.printStackTrace();
            }
            buf.unmap();
            sample.dispose();

            return FlowReturn.OK;
        });

        src.connect(new Element.PAD_ADDED() {
            @Override
            public void padAdded(Element element, Pad pad) {
                var s = pad.getCurrentCaps().getStructure(0);
                if (s.getName().equals("audio/x-raw")) {
                    pad.link(resample.getStaticPad("sink"));

                    pad.addDataProbe((pad1, buffer) -> {
                        buffer.setPresentationTimestamp(0);
                        return PadProbeReturn.OK;
                    });

                    src.disconnect(this);
                }
            }
        });

        pipeline.getBus().connect((Bus.EOS) source -> latch.countDown());

        pipeline.getBus().connect((Bus.INFO) (arg0, arg1, arg2) ->
            System.err.println("<reader:info> " + arg0.getName() + " [" + arg1 + "]: " + arg2));
        pipeline.getBus().connect((Bus.ERROR) (arg0, arg1, arg2) ->
            System.err.println("<reader:error> " + arg0.getName() + " [" + arg1 + "]: " + arg2));
//        pipeline.getBus().connect((Bus.MESSAGE) (arg0, arg1) ->
//            System.out.println("<reader:msg> " + arg1.getSource().getName() + " [" + arg1.getType() + "]: " + arg1.getStructure()));

        pipeline.addMany(src, resample, convert, sink);
        resample.link(convert);
        convert.link(sink);
    }

    private FloatBuffer await() {
        pipeline.play();

        try {
            latch.await();
        } catch (InterruptedException e) {
            e.printStackTrace();
            System.out.println("<reader> Interrupted!");
        }

        pipeline.stop();
        pipeline.close();

        return ByteBuffer.wrap(data.toByteArray()).asFloatBuffer();
    }
}
