package edu.bu.signstream.media.gstreamer;

import java.net.URI;
import java.time.Duration;
import java.util.*;
import java.util.concurrent.*;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import edu.bu.signstream.media.Audio;
import edu.bu.signstream.media.Media;
import org.freedesktop.gstreamer.*;
import org.freedesktop.gstreamer.event.SeekEvent;
import org.freedesktop.gstreamer.event.SeekFlags;
import org.freedesktop.gstreamer.event.SeekType;
import org.freedesktop.gstreamer.message.Message;
import org.freedesktop.gstreamer.message.MessageType;
import org.freedesktop.gstreamer.message.StateChangedMessage;

import edu.bu.signstream.media.MediaPlayer;
import edu.bu.signstream.media.Video;

import javax.swing.*;
import javax.swing.Timer;

public class GstMediaPlayer implements MediaPlayer {
    private static final int POLL_RATE = 30;

    Pipeline pipeline;

    private boolean prepared = false;
    private boolean playing = false;
    private boolean eos = false;

    private double rate = 1.0;
    private long end = -1;

    private final HashMap<Media, Long> offsets = new HashMap<>();
    private final ArrayList<GstMedia> media = new ArrayList<>();

    // TODO: private
    public int timestep = 1001;
    public int timebase = 30000;
    public double framerate = (double)timebase / (double)timestep;
    public Duration duration = Duration.ZERO;

    private final List<Consumer<Long>> posListeners = new ArrayList<>();
    private long pollLastPos = 0;
    private final Timer pollTimer = new Timer(POLL_RATE, e -> {
        if (this.playing()) {
            var p = pos();
            if (p != pollLastPos) {
                pollLastPos = p;
                for (var l : posListeners)
                    l.accept(p);
            }
        }
    });

    private final List<Consumer<Boolean>> stateListeners = new ArrayList<>();
    private final List<Consumer<Double>> rateListeners = new ArrayList<>();

    public GstMediaPlayer() {
        if (!Gst.isInitialized()) {
            System.out.println("Initializing GStreamer...");
            Gst.init();
            ElementFactory.listGetElements(ElementFactory.ListType.ANY, PluginFeature.Rank.NONE);
            System.out.println(Gst.getVersionString() + " initialized.");
        }

        System.out.println("Loaded plugins:");
        for (var p : Registry.get().getPluginList())
            System.out.println(" - " + p);

        pipeline = new Pipeline();

        pipeline.getBus().connect((Bus.INFO) (arg0, arg1, arg2) ->
            System.err.println("<main:info> " + arg0.getName() + " [" + arg1 + "]: " + arg2));
        pipeline.getBus().connect((Bus.ERROR) (arg0, arg1, arg2) ->
            System.err.println("<main:error> " + arg0.getName() + " [" + arg1 + "]: " + arg2));
//        pipeline.getBus().connect((Bus.MESSAGE) (arg0, arg1) ->
//            System.out.println("<main:msg> " + arg1.getSource().getName() + " [" + arg1.getType() + "]: " + arg1.getStructure()));

        pipeline.getBus().connect((Bus.STATE_CHANGED) (source, old, current, pending) -> {
            if (!prepared) return;

            if (source != pipeline) return;
            if (current != State.PLAYING && current != State.PAUSED) return;

            var state = current == State.PLAYING;
            var changed = playing != state;
            playing = state;

            if (playing)
                pollTimer.start();
            else
                pollTimer.stop();

            if (changed) {
                System.out.println("M: " + (playing ? "PLAYING" : "PAUSED"));
                for (var l : stateListeners)
                    l.accept(playing);
            }

            long pos;
            do {
                pos = pipeline.queryPosition(Format.TIME);
            } while (pos == -1);

            for (var l : posListeners)
                l.accept(pos);
        });

        Runnable onStopped = () -> {
            if (!prepared) return;

            System.out.println("M: EOS");

            eos = true;
            playing = false;

            for (var l : stateListeners)
                l.accept(false);
        };
        pipeline.getBus().connect((Bus.SEGMENT_DONE) (source, format, position) -> onStopped.run());
        pipeline.getBus().connect((Bus.EOS) source -> onStopped.run());
    }

    // Run a Callable in the GStreamer context. Should be used whenever interacting directly with the pipeline,
    // since certain methods are not thread safe.
    //
    // Be sure NOT to nest calls to run() or call run() from the GStreamer service thread,
    // since the executor is single threaded and will deadlock.
    private <V> V run(Callable<V> fn) {
        if (Thread.currentThread().getName().contains("gstreamer service thread"))
            throw new AssertionError("Call to GstMediaPlayer.run() from GStreamer service thread");

        try {
            return Gst.getExecutor().submit(fn).get(10, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            throw new AssertionError(e);
        } catch (ExecutionException e) {
            throw new RuntimeException(e);
        } catch (TimeoutException e) {
            JOptionPane.showMessageDialog(null, "DEADLOCK OCCURRED! Please send SignStreamOut.txt to the developers and restart SignStream!");
            System.err.println("====================================================================================");
            System.err.println("CRITICAL: GStreamer service thread deadlocked");
            for (Thread t : Thread.getAllStackTraces().keySet()) {
                if (t.getName().contains("gstreamer service thread")) {
                    System.err.println("THREAD: " + t.getName());
                    for (var el : t.getStackTrace()) {
                        System.err.println("    at " + el);
                    }
                }
            }
            System.err.println("====================================================================================");
            throw new RuntimeException("GStreamer task timed out");
        }
    }
    private void run(Runnable fn) {
        run(Executors.callable(fn));
    }

    @Override
    public List<Media> media() {
        return Collections.unmodifiableList(media);
    }

    @Override
    public List<Video> videos() {
        return media.stream()
            .filter(m -> m instanceof GstVideo)
            .map(v -> (Video)v)
            .collect(Collectors.toUnmodifiableList());
    }

    @Override
    public List<Audio> audio() {
        return media.stream()
            .filter(m -> m instanceof GstAudio)
            .map(v -> (Audio)v)
            .collect(Collectors.toUnmodifiableList());
    }

//    private Audio tempAudio;

    @Override
    public Video addVideo(URI uri) {
        System.out.println("M: ADD " + uri);

        var video = new GstVideo(uri, this);
        media.add(video);
        offset(video, 0);

//        video.view();

//        tempAudio = addAudio(uri);
//        tempAudio.view();

//        var start = System.currentTimeMillis();
//        var data = GstAudioReader.read(uri);
//        var end = System.currentTimeMillis();
//        System.out.println("Read " + data.remaining() + " floats at " + GstAudioReader.RATE + " in " + ((end - start) / 1000.0) + "s");

        return video;
    }

    @Override
    public Audio addAudio(URI uri) {
        System.out.println("addAudio(" + uri + ")");

        var audio = new GstAudio(uri, this);
        media.add(audio);
        offset(audio, 0);

        return audio;
    }

    @Override
    public void remove(Media m) {
        System.out.println("M: REMOVE " + m.uri());
        run(((GstMedia) m)::close);
        media.remove(m);
    }

    @Override
    public void offset(Media m, long offset) {
        offsets.put(m, offset);

        var max = media.stream()
            .map(o -> o.duration().toNanos() + offsets.get(o))
            .max(Long::compare)
            .orElseThrow();

        duration = Duration.ofNanos(max);
    }


    @Override
    public void prepare() {
        System.out.println("M: PREPARE START STATE=" + run(() -> pipeline.getState()));

        var latch = new CountDownLatch(1);
        pipeline.getBus().connect(new Bus.MESSAGE() {
            @Override
            public void busMessage(Bus bus, Message msg) {
                if (msg.getType() == MessageType.STATE_CHANGED) {
                    var state = ((StateChangedMessage) msg).getNewState();
                    if (msg.getSource() == pipeline && state == State.PLAYING) {
                        System.out.println("COUNT DOWN");
                        latch.countDown();
                        pipeline.getBus().disconnect(this);
                    }
                }
            }
        });

        System.out.println("M: PREPARE PLAY");
        play();
        try {
            System.out.println("M: PREPARE AWAIT");
            latch.await();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        System.out.println("M: PREPARE PAUSE");
        pause();

        System.out.println("M: PREPARE SEEK 0");
        seek(0);

        System.out.println("M: PREPARE COMPLETE STATE=" + run(() -> pipeline.getState()));
        prepared = true;
    }

    @Override
    public boolean prepared() {
        return prepared;
    }


    @Override
    public void play() {
        System.out.println("M: PLAY ");
        if (!playing())
            run(() -> pipeline.play());
    }

    @Override
    public void pause() {
        if (playing())
            run(() -> pipeline.pause());
    }

    @Override
    public boolean playing() {
        try {
            return run(() -> pipeline.isPlaying());
        } catch (Exception e) {
            throw new AssertionError(e);
        }
    }

    @Override
    public boolean stopped() {
        return eos;
    }


    @Override
    public void rate(double rate) {
        this.rate = Math.max(0.125, Math.min(8, Math.abs(rate)));
        System.out.println("M: RATE " + rate + " -> " + this.rate);

        for (var l : rateListeners)
            l.accept(this.rate);

        if (!eos) {
            var p = pos();
            System.out.println("M: RSEEK " + p);
            seek(SeekType.SET, p, SeekType.NONE, -1);
        }
    }

    @Override
    public double rate() {
        return rate;
    }


    @Override
    public void seek(long pos) {
        System.out.println("M: SEEK " + pos);
        if (pos >= duration.toNanos()) return;
        if (pos < 0) pos = 0;
        if (end != -1 && pos >= end) end = -1;

        seek(SeekType.SET, pos, SeekType.SET, end);
    }

    @Override
    public void seek(long start, long end) {
        System.out.println("M: SEEK " + start + " -> " + end);
        if (start >= duration.toNanos()) return;
        if (start < 0) start = 0;
        if (end > duration.toNanos()) end = -1;
        if (end != -1 && start >= end) end = -1;

        this.end = end;
        seek(SeekType.SET, start, SeekType.SET, end);
    }

    // Pipeline seek wrapper
    private void seek(SeekType ty0, long t0, SeekType ty1, long t1) {
        System.out.println("M: ISEEK " + ty0 + "/" + t0 + " -> " + ty1 + "/" + t1 + " E=" + end);
        var flags = EnumSet.of(SeekFlags.FLUSH, SeekFlags.ACCURATE, SeekFlags.TRICKMODE);
        for (var m : media) {
            var offset = offsets.get(m);
            run(() -> {
                if (t0 - offset < 0) {
                    m.sink().set("ts-offset", offset - t0);
                    m.sink().sendEvent(new SeekEvent(rate, Format.TIME, flags, ty0, 0, ty1, t1));
                } else {
                    m.sink().set("ts-offset", 0);
                    m.sink().sendEvent(new SeekEvent(rate, Format.TIME, flags, ty0, t0 - offset, ty1, t1));
                }
            });
        }
    }



    @Override
    public long pos() {
        return run(() -> {
            long pos;
            do {
                pos = pipeline.queryPosition(Format.TIME);
            } while (pos == -1);

            return pos;
        });
    }

    @Override
    public Duration duration() {
        return duration;
    }


    // Attach a callback for pos updates.
    // During playback, the pos is polled every `POLL_RATE` ms.
    @Override
    public void addPosListener(Consumer<Long> fn) {
        posListeners.add(fn);
    }
    @Override
    public void removePosListener(Consumer<Long> fn) {
        posListeners.remove(fn);
    }

    // Attach a callback for play/pause updates.
    // Param true if playing.
    @Override
    public void addStateListener(Consumer<Boolean> fn) {
        stateListeners.add(fn);
    }
    @Override
    public void removeStateListener(Consumer<Boolean> fn) {
        stateListeners.remove(fn);
    }

    @Override
    public void addRateListener(Consumer<Double> fn) {
        rateListeners.add(fn);
    }
    @Override
    public void removeRateListener(Consumer<Double> fn) {
        rateListeners.remove(fn);
    }

    @Override
    public void close() {
        prepared = false;
        pollTimer.stop();

        // Clone before iterating because `remove` modifies the array at the same time
        for (var m : new ArrayList<>(media))
            remove(m);

        pipeline.close();
    }


    //
    // Temp for quick replacement. Remove later, should not be used.
    //


    @Deprecated
    public long ts() {
        return tsFromPos(pos());
    }

    // Current frame
    @Deprecated
    public long frame() {
        double sec = pos() / 1000000000.0;
        return Math.round(sec * framerate);
    }

    // Pos from ts
    @Deprecated
    public long posFromTs(long ts) {
        double frame = ts / (double)timestep;
        double sec = frame / framerate;
        return Math.round(sec * 1000000000.0);
    }

    // Current ts
    @Deprecated
    public long tsFromPos(long pos) {
        double sec = pos / 1000000000.0;
        long frame = Math.round(sec * framerate);
        return frame * timestep;
    }

    @Deprecated
    public long posFromFrame(long frame) {
        double sec = (double)frame / framerate;
        return Math.round(sec * 1000000000.0);
    }

    @Deprecated
    public long frameFromPos(long pos) {
        double sec = pos / 1000000000.0;
        return Math.round(sec * framerate);
    }

    @Deprecated
    public double frFromPos(long pos) {
        return (double)pos / (double)duration.toNanos();
    }
}
