package edu.bu.signstream.media.gstreamer;


import java.awt.*;
import java.net.URI;
import java.time.Duration;

import javax.swing.JComponent;

import edu.bu.signstream.media.MediaPlayer;
import edu.bu.signstream.media.Video;
import edu.bu.signstream.media.ui.VideoView;
import org.freedesktop.gstreamer.Element;
import org.freedesktop.gstreamer.ElementFactory;
import org.freedesktop.gstreamer.Pad;
import org.freedesktop.gstreamer.PadLinkException;
import org.freedesktop.gstreamer.elements.AppSink;

public class GstVideo implements GstMedia, Video {
    private static final String FORMAT = "video/x-raw";
    private static final String ID_PREFIX = "V";

    private static int ID = 0;

    private final URI uri;

    private int width;
    private int height;
    private double framerate;
    private Duration duration;

    private int timestep;
    private int timebase;

    private final GstMediaPlayer player;

    private final Element decode;
    private final Element convert;
    private final Element sink;

    private final GstVideoImage image;

    public GstVideo(URI uri, GstMediaPlayer player) {
        this.player = player;
        this.uri = uri;

        var id = ID_PREFIX + ID;

        System.out.println("Probing video " + uri.toString());
        GstProbe.probe(uri, FORMAT, (caps, duration) -> {
            this.duration = duration;
            width = caps.getInteger("width");
            height = caps.getInteger("height");
            var fr = caps.getFraction("framerate");
            timestep = fr.getDenominator();
            timebase = fr.getNumerator();
            framerate = (double) timebase / (double) timestep;
        });
        System.out.println("Probing finished!");

        // Set up the elements for the real pipeline
        decode = ElementFactory.make("uridecodebin", "decode" + id);
        decode.set("uri", uri.toString());
        decode.connect(new Element.PAD_ADDED() {
            @Override
            public void padAdded(Element e, Pad pad) {
                System.out.println(uri + ": " + pad.getName() + " [" + pad.getTypeName() + "]");

                if (pad.getName().equals("src_0")) {
                    try {
                        pad.link(convert.getStaticPad("sink"));
                    } catch (PadLinkException ex) {
                        ex.printStackTrace();
                    }
                }
            }
        });
        convert = ElementFactory.make("videoconvert", "convert" + id);

        var appsink = new AppSink("sink" + ID_PREFIX + id);
        image = new GstVideoImage(appsink);
        image.setPreferredSize(new Dimension(width, height));
        sink = appsink;

        player.pipeline.addMany(decode, convert, sink);
        convert.link(sink);

        System.out.println("Done loading video " + id);

        ID++;
    }

    @Override
    public URI uri() {
        return uri;
    }

    @Override
    public int width() {
        return width;
    }

    @Override
    public int height() {
        return height;
    }

    @Override
    public double framerate() {
        return framerate;
    }

    @Override
    public Duration duration() {
        return duration;
    }

    @Override
    public int timestep() {
        return timestep;
    }

    @Override
    public int timebase() {
        return timebase;
    }

    @Override
    public JComponent image() {
        return image;
    }

    @Override
    public VideoView view() {
        return new VideoView(player, this, image);
    }

    @Override
    public Element sink() {
        return sink;
    }

    @Override
    public void close() {
        player.pipeline.remove(decode);
        player.pipeline.remove(convert);
        player.pipeline.remove(sink);
    }
}
