package edu.bu.signstream.media.ui;

import edu.bu.signstream.media.Media;
import edu.bu.signstream.media.MediaPlayer;
import edu.bu.signstream.media.ui.icon.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.function.Consumer;

public abstract class MediaView extends JFrame {
    private final ToggleButton play = new ToggleButton(b -> b ? new PauseIcon(true) : new PlayIcon(true), false);
    private final JButton slower = new ToggleButton(SpeedSlowerIcon::new);
    private final JButton faster = new ToggleButton(SpeedFasterIcon::new);

    private final JButton forward1 = new ToggleButton(GoFrameForward1Icon::new);
    private final JButton back1 = new ToggleButton(GoFrameBack1Icon::new);
    private final JButton forward2 = new ToggleButton(GoFrameForward2Icon::new);
    private final JButton back2 = new ToggleButton(GoFrameBack2Icon::new);
    private final JButton forward3 = new ToggleButton(GoFrameForward3Icon::new);
    private final JButton back3 = new ToggleButton(GoFrameBack3Icon::new);
    private final JButton start = new ToggleButton(GoFirstFrameIcon::new);
    private final JButton end = new ToggleButton(GoLastFrameIcon::new);

    protected static final int SLIDER_MAX = 1000000;
    private final JSlider slider = new JSlider(JSlider.HORIZONTAL, SLIDER_MAX);

    private final JLabel rate = new JLabel();

    private final Consumer<Long> posListener;
    private final Consumer<Double> rateListener;
    private final Consumer<Boolean> stateListener;

    protected MediaPlayer player;
    protected Media media;

    public MediaView(MediaPlayer player, Media media) {
        this.player = player;
        this.media = media;

        posListener = p -> SwingUtilities.invokeLater(() -> slider.setValue((int)((double)p / player.duration().toNanos() * SLIDER_MAX)));
        rateListener = r -> SwingUtilities.invokeLater(() -> rate.setText(r == 1 ? "" : "X" + r));
        stateListener = s -> SwingUtilities.invokeLater(() -> play.set(s));

        player.addPosListener(posListener);
        player.addRateListener(rateListener);
        player.addStateListener(stateListener);

        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                super.windowClosing(e);
                player.removePosListener(posListener);
                player.removeRateListener(rateListener);
                player.removeStateListener(stateListener);
            }
        });
    }

    protected abstract JComponent view();
    protected abstract JPanel controls(JPanel panel);

    protected void build() {
        var playPauseGroup = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        playPauseGroup.setAlignmentX(Component.LEFT_ALIGNMENT);
        playPauseGroup.setBackground(Color.BLACK);
        slower.addActionListener(l -> { if (player.rate() >= 0.5) player.rate(player.rate() / 2); });
        playPauseGroup.add(slower);
        play.addActionListener(l -> { if (play.state()) player.pause(); else player.play(); });
        playPauseGroup.add(play);
        faster.addActionListener(l -> { if (player.rate() <= 4.0) player.rate(player.rate() * 2); });
        playPauseGroup.add(faster);

        var controlsGroup = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        controlsGroup.setAlignmentX(Component.RIGHT_ALIGNMENT);
        controlsGroup.setBackground(Color.BLACK);
        controlsGroup = controls(controlsGroup);

        var stepGroup = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        stepGroup.setBackground(Color.BLACK);
        start.addActionListener(l -> player.seek(0));
        stepGroup.add(start);
        back3.addActionListener(l -> player.seek(player.pos() - 1000000000));
        stepGroup.add(back3);
        back2.addActionListener(l -> player.seek(player.pos() - 500000000));
        stepGroup.add(back2);
        stepGroup.add(back1);
        stepGroup.add(forward1);
        forward2.addActionListener(l -> player.seek(player.pos() + 500000000));
        stepGroup.add(forward2);
        forward3.addActionListener(l -> player.seek(player.pos() + 1000000000));
        stepGroup.add(forward3);
        end.addActionListener(l -> player.seek(player.duration().toNanos() - 1000000000));
        stepGroup.add(end);

        var strip = new JPanel();
        strip.setLayout(new BoxLayout(strip, BoxLayout.X_AXIS));
        strip.setBackground(Color.BLACK);
        strip.setMinimumSize(new Dimension(300, 22));
        strip.setPreferredSize(new Dimension(300, 22));
        strip.setMaximumSize(new Dimension(Integer.MAX_VALUE, 22));

        strip.add(playPauseGroup);
        strip.add(Box.createHorizontalGlue());
        strip.add(stepGroup);
        if (controlsGroup != null) {
            strip.add(Box.createHorizontalGlue());
            strip.add(controlsGroup);
        }

        var container = new JPanel();
        container.setLayout(new BoxLayout(container, BoxLayout.Y_AXIS));
        container.setBorder(BorderFactory.createLineBorder(Color.BLUE));
        container.add(view());
        container.add(strip);
        container.add(slider);
        getContentPane().add(container);

        var glass = (JPanel)getGlassPane();
        glass.setLayout(new FlowLayout(FlowLayout.LEFT));
        rate.setFont(new Font(rate.getFont().getName(), Font.PLAIN, 22));
        rate.setForeground(Color.white);
        glass.add(rate);
        glass.setVisible(true);

        setTitle(media.uri().toString());
        setMinimumSize(new Dimension(300, 200));

        pack();
        setVisible(true);
    }
}
